/* Richard Liang
   June 7, 2003
   
   Code to set up a 3D percolation cluster.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "3dhf.h"


// A quick and dirty way to generate random floats.
#define fran() ((float) random()/(float) 0x7fffffff)


// A function to allocate a new percCluster3D, empty and 
// uninitialized.
percCluster3D_t* create_percCluster3D_t(int size)
{
  // Counters, etc.
  int i,j;

  // We start by creating a new one dynamically, then 
  // returning a pointer to it.
  percCluster3D_t* newCluster = (percCluster3D_t*)malloc
    (sizeof(percCluster3D_t));

  // Now we go in and allocate the arrays.
  // Note the sizes:

  // adj shows how many other points connect to each point 
  //  in the grid.
  //  Thus its size is size*size*size
  newCluster->adj = (char***)malloc(sizeof(char**) * size);
  for (i=0; i<size; i++)
    {
      (newCluster->adj)[i] = (char**)malloc(sizeof(char*) * size);
      
      for (j=0; j<size; j++)
	{
	  (newCluster->adj)[i][j] = (char*)malloc(sizeof(char)
						  * size);
	}
    }

  // x reflects a bond pointing out in the positive x direction
  //  from any given point in the grid.
  //  Thus its size is (size-1)*size*size
  newCluster->x = (char***)malloc(sizeof(char**) * (size-1));
  for (i=0; i<size-1; i++)
    {
      (newCluster->x)[i] = (char**)malloc(sizeof(char*) * size);

      for (j=0; j<size; j++)
	{
	  (newCluster->x)[i][j] = (char*)malloc(sizeof(char)
						* size);
	}
    }

  // y shows bonds pointing in the positive y direction from
  //  any given point in the grid.
  //  Thus its size is size*(size-1)*size
  newCluster->y = (char***)malloc(sizeof(char**) * (size));
  for (i=0; i<size; i++)
    {
      (newCluster->y)[i] = (char**)malloc(sizeof(char*) * 
					  (size-1));
      for (j=0; j<size-1; j++)
	{
	  (newCluster->y)[i][j] = (char*)malloc(sizeof(char)
						* size);
	}
    }


  // z shows bonds pointing in the positive z direction from
  //  any given point in the grid.
  //  Thus its size is size*size*(size-1)
  newCluster->z = (char***)malloc(sizeof(char**) * (size));
  for (i=0; i<size; i++)
    {
      (newCluster->z)[i] = (char**)malloc(sizeof(char*) * size);
      for (j=0; j<size; j++)
	{
	  (newCluster->z)[i][j] = (char*)malloc(sizeof(char)
						* (size-1));
	}
    }

  // Initialize everything else.
  newCluster->size = size;
  newCluster->p = -1;
  newCluster->rseed = -1;
  newCluster->trialNo = -1;

  return newCluster;
}

// A function to dispose of a percolation cluster.
void dispose_percCluster3D_t(percCluster3D_t* cluster)
{
  // counters, etc.
  int i,j;
  int size = cluster->size;

  // Dispose of the arrays, starting with adj.
  for (i=0; i<size; i++)
    {
      for (j=0; j<size; j++)
	{
	  free((cluster->adj)[i][j]);
	}
      free((cluster->adj)[i]);
    }
  free(cluster->adj);

  // Now x.
  for (i=0; i<(size-1); i++)
    {
      for (j=0; j<size; j++)
	{
	  free((cluster->x)[i][j]);
	}
      free((cluster->x)[i]);
    }
  free(cluster->x);

  // Now y.
  for (i=0; i<size; i++)
    {
      for (j=0; j<(size-1); j++)
	{
	  free((cluster->y)[i][j]);
	}
      free((cluster->y)[i]);
    }
  free(cluster->y);

  // Now z.
  for (i=0; i<size; i++)
    {
      for (j=0; j<size; j++)
	{
	  free((cluster->z)[i][j]);
	}
      free((cluster->z)[i]);
    }
  free(cluster->z);

  // Free the rest of it.
  free(cluster);

  cluster = NULL;
}


// A function to actually set up a percolation cluster.
// Returns information about the largest cluster.
clusterInfo_t doPerc(percCluster3D_t* myCluster, double p,
		     int rseed, int trialNo)
{
  // Counters, etc.
  int i,j,k;
  int thirdDigit;
  int secondDigit;
  //int min;
  
  // An array that allows us to find the largest cluster,
  // by a method copied from Dr. Barlow's code.
  int ***findCluster;
  int *clusterStats;
  int totalElements;

  int numSweepsCompleted = -1;
  int success = -1;

  int currElement;

  clusterInfo_t myInfo;
  int maxClusterNo;

  int size;

  // Start by getting the size of the percolation cluster.
  size = myCluster->size;

  // Fill in the requisite info in the percolation cluster.
  myCluster->p = p;
  myCluster->rseed = rseed;
  myCluster->trialNo = trialNo;

  // Now create bonds.  Note the dimensions and the way
  // we swap around i,j,k for each array.
  for (i=0; i<size; i++)
    {
      for (j=0; j<size; j++)
	{
	  for (k=0; k<size-1; k++)
	    {
	      (myCluster->x)[k][i][j] = (char)(fran() < p);
	      (myCluster->y)[i][k][j] = (char)(fran() < p);
	      (myCluster->z)[i][j][k] = (char)(fran() < p);
	    }
	}
    }


  // Now it's time to find the largest cluster.  We fill a 3D
  // array lexicographically (first x, then y, then z).
  findCluster = (int***)malloc(sizeof(int**) * size);
  for (i=0; i<size; i++)
    {
      findCluster[i] = (int**)malloc(sizeof(int*) * size);
      for (j=0; j<size; j++)
	{
	  findCluster[i][j] = (int*)malloc(sizeof(int) * size);
	}
    }

  // Note: this might not be the most intuitive way of 
  // filling the array.  
  for (i=0; i<size; i++)
    {
      thirdDigit = (size*size)*i;
      for (j=0; j<size; j++)
	{
	  secondDigit = size*j;
	  for (k=0; k<size; k++)
	    {
	      findCluster[i][j][k] = thirdDigit +
		secondDigit + k;
	    }
	}
    }

  // Now, we go through and find the largest cluster.
  // success indicates whether there has been a change
  // in the array during the last sweep.
  success = 0;
  numSweepsCompleted = 0;
  
  while (0 == success)
    {
      success = 1;

      for (i=0; i<size; i++)
	{
	  for (j=0; j<size; j++)
	    {
	      for (k=0; k<size-1; k++)
		{
		  // First, we examine x.
		  if (1 == (myCluster->x)[k][i][j])
		    {
		      // There is a bond between 
		      // (k,i,j) and (k+1,i,j).
		      // Set the corresponding values 
		      // in findCluster to their minimum.
		      if (findCluster[k][i][j] !=
			  findCluster[k+1][i][j])
			{
			  // Indicate a change in this sweep.
			  success = 0;
			  if (findCluster[k][i][j] < 
			      findCluster[k+1][i][j])
			    findCluster[k+1][i][j] = 
			      findCluster[k][i][j];
			  else
			    findCluster[k][i][j] = 
			      findCluster[k+1][i][j];
			} 
		    }
		  
		  // Next we examine y.
		  if (1 == (myCluster->y)[i][k][j])
		    {
		      // There is a bond between 
		      // (i,k,j) and (i,k+1,j).
		      // Set the corresponding values 
		      // in findCluster to their minimum.
		      if (findCluster[i][k][j] !=
			  findCluster[i][k+1][j])
			{
			  // Indicate a change in this sweep.
			  success = 0;
			  if (findCluster[i][k][j] < 
			      findCluster[i][k+1][j])
			    findCluster[i][k+1][j] = 
			      findCluster[i][k][j];
			  else
			    findCluster[i][k][j] = 
			      findCluster[i][k+1][j];
			} 
		    }
		  
		  // Finally we look at z.
		  if (1 == (myCluster->z)[i][j][k])
		    {
		      // There is a bond between 
		      // (i,j,k) and (i,j,k+1).
		      // Set the corresponding values 
		      // in findCluster to their minimum.
		      if (findCluster[i][j][k] !=
			  findCluster[i][j][k+1])
			{
			  // Indicate a change in this sweep.
			  success = 0;
			  if (findCluster[i][j][k] < 
			      findCluster[i][j][k+1])
			    findCluster[i][j][k+1] = 
			      findCluster[i][j][k];
			  else
			    findCluster[i][j][k] = 
			      findCluster[i][j][k+1];
			} 
		    }
		}
	    }
	}
      numSweepsCompleted++;
    }
  
  // Print out some status report.
  printf("Maximum cluster found in %d sweeps.\n",
	 numSweepsCompleted);

  
  // Now we identify the maximum cluster.  Allocate clusterStats
  // accordingly.
  totalElements = size*size*size;
  clusterStats = (int*)malloc(sizeof(int) * totalElements);
  memset(clusterStats,0,sizeof(int)*totalElements);

  for (i=0; i<size; i++)
    {
      for (j=0; j<size; j++)
	{
	  for (k=0; k<size; k++)
	    {
	      currElement = findCluster[i][j][k];
	      clusterStats[currElement]++;
	    }
	}
    } 
  
  maxClusterNo = 0;
  for (i=0; i<totalElements; i++)
    {
      if (clusterStats[i] > clusterStats[maxClusterNo])
	maxClusterNo = i;
    }

  // We make our clusterInfo_t.
  myInfo.clusterNo = maxClusterNo;
  myInfo.clusterSize = clusterStats[maxClusterNo];

  // We make a graph structure out of the maximum cluster.
  // Also, for the bonds that are not part of the maximum cluster,
  // we give it a negative value in adj.
  for (i=0; i<size; i++)
    {
      for (j=0; j<size; j++)
	{
	  for (k=0; k<size; k++)
	    {
	      (myCluster->adj)[i][j][k] = 0;
	      
	      if (findCluster[i][j][k] == maxClusterNo)
		{
		  if (i>0 && (myCluster->x)[i-1][j][k]>0)
		    (myCluster->adj)[i][j][k]++;
		  if (j>0 && (myCluster->y)[i][j-1][k]>0)
		    (myCluster->adj)[i][j][k]++;
		  if (k>0 && (myCluster->z)[i][j][k-1]>0)
		    (myCluster->adj)[i][j][k]++;
		  if (i<size-1 && (myCluster->x)[i][j][k]>0)
		    (myCluster->adj)[i][j][k]++;
		  if (j<size-1 && (myCluster->y)[i][j][k]>0)
		    (myCluster->adj)[i][j][k]++;
		  if (k<size-1 && (myCluster->z)[i][j][k]>0)
		    (myCluster->adj)[i][j][k]++;
		}
	      else
		{
		  if (i>0 && (myCluster->x)[i-1][j][k]>0)
		    (myCluster->adj)[i][j][k]--;
		  if (j>0 && (myCluster->y)[i][j-1][k]>0)
		    (myCluster->adj)[i][j][k]--;
		  if (k>0 && (myCluster->z)[i][j][k-1]>0)
		    (myCluster->adj)[i][j][k]--;
		  if (i<size-1 && (myCluster->x)[i][j][k]>0)
		    (myCluster->adj)[i][j][k]--;
		  if (j<size-1 && (myCluster->y)[i][j][k]>0)
		    (myCluster->adj)[i][j][k]--;
		  if (k<size-1 && (myCluster->z)[i][j][k]>0)
		    (myCluster->adj)[i][j][k]--;
		}
	    }
	}
    }


  // We're done.  Free all our memory.
  for (i=0; i<size; i++)
    {
      for (j=0; j<size; j++)
	{
	  free(findCluster[i][j]);
	}
      free(findCluster[i]);
    }
  free(findCluster);

  free(clusterStats);

  return myInfo;
}
