/* Richard Liang
   Routines that take a percolation cluster and draws it.

   June 10, 2003
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <GL/glut.h>
#include "3dhf.h"
#include "showCluster.h"

// Global variables (ick) so that we can enable the display
// function to see the percolation cluster.
percCluster3D_t *cluster;

// And more shared variables to do with manipulating the
// cluster so we can play with it.
// these variables rotate in the x, y, and z axes respectively.
double theta;
double phi;
double psi;
double spinAmount;


// Indicates if we set the thing on autopilot twisting.
int twisting;
// The threshold for the twisting: how many degrees we go either way.
int twistThreshold;
// The current twist positions for all three axes.
double currTwistX, currTwistY, currTwistZ;
// How fast we twist.
double twistAmount;
// Which directions we are currently twisting (clockwise or counterclockwise)
int twistModeX, twistModeY, twistModeZ;
// Which axis we are currently twisting in (0~x, 1~y, 2~z)
int twistAxis;


// These variables translate the cluster.
double moveX;
double moveY;
double moveZ;
double moveAmount;

// This variable scales the cluster.
double scale;
double scaleAmount;

// This variable indicates how much we are going to show:
// just the largest cluster (0), or everything (1).
int toShow;

// This variable sets your position.
double myPositionX;
double myPositionY;
double myPositionZ;

// This variable determines whether we want to use
// an orthogonal or perspective projection matrix.
// 0 for perspective, 1 for orthogonal.
int myProjection;

// Particulars for visual tuning.
float bondThickness;
float nodeSize;
int lightStyle;
GLfloat light_position[4];

// Axis markers.
float axisMarkerSize;
float axisLineSize;
GLfloat XAxis_diffuse[4];
GLfloat YAxis_diffuse[4];
GLfloat ZAxis_diffuse[4];

// Indicates whether or not we are going to show the 
// axes of the absolute coordinate system.
int showAbsAxes;

// Indicates whether or not we are going to show the
// axes of the coordinate system local to the cluster.
int showLocalAxes;

// The colours we use.
GLfloat prim_node_diffuse[4];
GLfloat prim_node_specular[4];
GLfloat prim_node_shininess[1];

GLfloat sec_node_diffuse[4];
GLfloat sec_node_specular[4];
GLfloat sec_node_shininess[1];

GLfloat prim_bond_diffuse[4];
GLfloat prim_bond_specular[4];
GLfloat prim_bond_shininess[1];

GLfloat sec_bond_diffuse[4];
GLfloat sec_bond_specular[4];
GLfloat sec_bond_shininess[1];


// The name of our config file.
char configfile[50] = "3dperc.conf";


// This function does the drawing... It is like the "main"
// function specified in the book.
void drawCluster(percCluster3D_t *myCluster,
		 int *argc, char **argv)
{
  // Mostly this is lifted from examples in the book.
  glutInit(argc, argv);
  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
  glutInitWindowSize(500,500);
  glutInitWindowPosition(50,50);
  glutCreateWindow("My percolation cluster");

  // Here we set up the clear colour, etc.
  cluster = myCluster;
  init();

  glutDisplayFunc(display);
  glutKeyboardFunc(keyboard);
  glutReshapeFunc(reshape);
  glutMainLoop();
}

// The init function.  This sets all the things we only set
// once during the lifetime of the display.
void init()
{
  // Information regarding the light: we set one light, 
  // right where we are, as if we are using a flashlight.
  // This is mostly copied from p.180 of the book.
  //GLfloat light_position[] = {0.0, 0.0, myPosition, 1.0};
  GLfloat white_light[] = {1.0, 1.0, 1.0, 1.0};
  GLfloat lmodel_ambient[] = {0.1, 0.1, 0.1, 1.0};

  // Set the clear color.
  glClearColor(0.0, 0.0, 0.0, 0.0);
  // We don't want any fancy-pants colouring: all polygons
  // will be uniformly coloured.
  glShadeModel(GL_FLAT);

  theta = 0.0;
  phi = 0.0;
  psi = 0.0;
  
  moveX = 0.0;
  moveY = 0.0;
  moveZ = 0.0;

  scale = 1.0;

  toShow = 0;

  // Twisting stuff.
  twisting = 0;
  currTwistX = currTwistY = currTwistZ = 0.0;
  twistModeX = twistModeY = twistModeZ = 0;
  twistAxis = 0;
  
  showAbsAxes = showLocalAxes = 1;

  // Read some preferences from our .conf file.
  config();

  // Indicate that initially we are to use a perspective 
  // projection matrix.
  myProjection = 0;

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  // Set lighting up.  We set all material things to have
  // the same properties.
  //glLightfv(GL_LIGHT0, GL_POSITION, light_position);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, white_light);
  glLightfv(GL_LIGHT0, GL_SPECULAR, white_light);
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);

  // We want to attenuate the light at a distance.
  //glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION, 1.5);
  //glLightf(GL_LIGHT0, GL_LINEAR_ATTENUATION, 0.5);
  //glLightf(GL_LIGHT0, GL_QUADRATIC_ATTENUATION, 0.2);

  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glEnable(GL_DEPTH_TEST);


  // Things seem to get darker as we scale.  This should fix that.
  //glEnable(GL_RESCALE_NORMAL);
  // That worked, but somehow the z-connectors weren't the right
  // colour.  This fixes it (why?).
  glEnable(GL_NORMALIZE);


  // Prettying things up a bit more with antialiasing.
  // Most of this is taken from the book.
  //glEnable(GL_POLYGON_SMOOTH);
  //glEnable(GL_BLEND);
  //glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  //glHint(GL_POLYGON_SMOOTH_HINT, GL_DONT_CARE);

}


// A function to read configuration settings from a configuration file.
// Note: don't be fooled, this is a very simple function and all it
// does is read in settings from a file.  Unfortunately it's also
// quite a long function.
void config()
{
  // The size of the cluster.
  double size = (double)(cluster->size);

  // A filehandle.
  FILE *settings;

  // Things we will set.  These are the default values.
  double newBondThickness = 0.10;
  double newNodeSize = 0.12;
  double newSpinAmount = 1.0;
  double newMoveAmount = 0.01;
  double newScaleAmount = 1.1;  

  double newAxisMarkerSize = 0.2;
  double newAxisLineSize = 0.015;
  GLfloat new_XAxis_diffuse[] = {1.0, 0.0, 0.0, 1.0};
  GLfloat new_YAxis_diffuse[] = {0.0, 1.0, 0.0, 1.0};
  GLfloat new_ZAxis_diffuse[] = {0.0, 0.0, 1.0, 1.0};

  // Camera.
  double newPositionX = 1.2*size;
  double newPositionY = newPositionX;
  double newPositionZ = newPositionX;

  
  // Lights.
  int newLightStyle = 1;
  GLfloat new_light_position[4] = {newPositionX, newPositionY, newPositionZ,
				   (double)newLightStyle};

  
  // Twisting stuff.
  int newTwistThreshold = 40;
  double newTwistAmount = 1.0;


  // Colours.
  GLfloat new_PN_diffuse[4] = {1.0, 1.0, 1.0, 1.0};
  GLfloat new_PN_specular[4] = {1.0, 1.0, 1.0, 1.0};
  GLfloat new_PN_shininess[1] = {120.0};

  GLfloat new_SN_diffuse[] = {1.0, 0.0, 0.0, 1.0};
  GLfloat new_SN_specular[] = {1.0, 1.0, 1.0, 1.0};
  GLfloat new_SN_shininess[] = { 120.0 };

  GLfloat new_PB_diffuse[] = {0.0, 1.0, 0.0, 1.0};
  GLfloat new_PB_specular[] = {1.0, 1.0, 1.0, 1.0};
  GLfloat new_PB_shininess[] = { 120.0 };

  GLfloat new_SB_diffuse[] = {0.0, 0.0, 1.0, 1.0};
  GLfloat new_SB_specular[] = {1.0, 1.0, 1.0, 1.0};
  GLfloat new_SB_shininess[] = { 120.0 };

  int n;

  char parameter[50] = "";
  double newvalue = -1;
  char del[10] = "";

  // Open the file.
  if (NULL == (settings = fopen(configfile,"r")))
    {
      printf("Could not open the config file.  Using default settings.\n\n");
    }
  else
    {
      // Start reading.
      while ((3 == (n = fscanf(settings,"%s = %lf %1s",parameter,&newvalue,del))) 
	     && (0 == feof(settings)) && (0 == ferror(settings)) 
	     && (del[0] == 'd'))
	{
	  if (0 == strcmp(parameter,"BONDTHICKNESS"))
	    {
	      // We are trying to set the bond thickness.  Check
	      // that it is positive.
	      if (0 >= newvalue)
		{
		  printf("Bond thickness must be positive.\n");
		  printf("Not changing the bond thickness.\n");
		}
	      else
		{
		  newBondThickness = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"NODESIZE"))
	    {
	      // We are trying to set node size.  Check that it is positive.
	      if (0 >= newvalue)
		{
		  printf("Node size must be positive.\n");
		  printf("Not changing the node size.\n");
		}
	      else
		{
		  newNodeSize = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"AXISMARKERSIZE"))
	    {
	      // We are trying to set axis marker size.  
	      // Check that it is not negative.
	      if (0 > newvalue)
		{
		  printf("Axis marker size must be non-negative.\n");
		  printf("Not changing the axis marker size.\n");
		}
	      else
		{
		  newAxisMarkerSize = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"AXISLINESIZE"))
	    {
	      // We are trying to set axis line size.  
	      // Check that it is non-negative.
	      if (0 > newvalue)
		{
		  printf("Axis line size must be non-negative.\n");
		  printf("Not changing the axis line size.\n");
		}
	      else
		{
		  newAxisLineSize = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"MOVEAMOUNT"))
	    {
	      // We are trying to set the move amount.  Check that it is positive.
	      if (0 >= newvalue)
		{
		  printf("Move amount must be positive.\n");
		  printf("Not changing move amount.\n");
		}
	      else
		{
		  newMoveAmount = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"SPINAMOUNT"))
	    {
	      // We are trying to set the spin amount.  Check that it is positive.
	      if (0 >= newvalue)
		{
		  printf("Spin amount must be positive.\n");
		  printf("Not changing spin amount.\n");
		}
	      else
		{
		  newSpinAmount = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"SCALEAMOUNT"))
	    {
	      // We are trying to set the scale amount.  
	      // Check that it is greater than 1.
	      if (1 >= newvalue)
		{
		  printf("Scale amount must be greater than 1.\n");
		  printf("Not changing scale amount.\n");
		}
	      else
		{
		  newScaleAmount = newvalue;
		}
	    }

	  // Twisting stuff.
	  else if (0 == strcmp(parameter,"TWISTAMOUNT"))
	    {
	      // We are trying to set the twist amount.  
	      // Check that it is positive.
	      if (0 >= newvalue)
		{
		  printf("Twist amount must be positive.\n");
		  printf("Not changing twist amount.\n");
		}
	      else
		{
		  newTwistAmount = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"THRESHOLD"))
	    {
	      // We are trying to set the threshold.  
	      // Check that it is greater than 0.
	      if (0 >= newvalue)
		{
		  printf("Twist threshold must be greater than 1.\n");
		  printf("Not changing twist threshold.\n");
		}
	      else
		{
		  newTwistThreshold = (int)newvalue;
		}
	    }





	  // Camera settings.
	  else if (0 == strcmp(parameter,"CAMERAX"))
	    {
	      newPositionX = newvalue * size;
	    }
	  else if (0 == strcmp(parameter,"CAMERAY"))
	    {
	      newPositionY = newvalue * size;
	    }
	  else if (0 == strcmp(parameter,"CAMERAZ"))
	    {
	      newPositionZ = newvalue * size;
	    }

	  // Light settings.
	  else if (0 == strcmp(parameter,"LIGHTX"))
	    {
	      new_light_position[0] = newvalue * size;
	    }
	  else if (0 == strcmp(parameter,"LIGHTY"))
	    {
	      new_light_position[1] = newvalue * size;
	    }
	  else if (0 == strcmp(parameter,"LIGHTZ"))
	    {
	      new_light_position[2] = newvalue * size;
	    }
	    

	  // Colour settings for the primary nodes.
	  else if (0 == strcmp(parameter,"PNDRED"))
	    {
	      // We are trying to set the redness of the primary nodes.
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing PNDRED.\n");
		}
	      else
		{
		  new_PN_diffuse[0] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"PNDGREEN"))
	    {
	      // We are trying to set the greenness of the primary nodes.
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing PNDGREEN.\n");
		}
	      else
		{
		  new_PN_diffuse[1] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"PNDBLUE"))
	    {
	      // We are trying to set the blueness of the primary nodes.
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing PNDBLUE.\n");
		}
	      else
		{
		  new_PN_diffuse[2] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"PNDALPHA"))
	    {
	      // We are trying to set the alpha of the primary nodes.
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing PNDALPHA.\n");
		}
	      else
		{
		  new_PN_diffuse[3] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"PNSRED"))
	    {
	      // We are trying to set the redness of the primary nodes.
	      // (specular)
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing PNSRED.\n");
		}
	      else
		{
		  new_PN_specular[0] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"PNSGREEN"))
	    {
	      // We are trying to set the greenness of the primary nodes.
	      // (specular)
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing PNSGREEN.\n");
		}
	      else
		{
		  new_PN_specular[1] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"PNSBLUE"))
	    {
	      // We are trying to set the blueness of the primary nodes.
	      // (specular)
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing PNSBLUE.\n");
		}
	      else
		{
		  new_PN_specular[2] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"PNSALPHA"))
	    {
	      // We are trying to set the alpha of the primary nodes.
	      // (specular)
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing PNSALPHA.\n");
		}
	      else
		{
		  new_PN_specular[3] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"PNSHINE"))
	    {
	      // We are trying to set the shininess of the primary nodes.
	      // (specular)
	      // Check that it is positive.
	      if (0 >= newvalue)
		{
		  printf("Shininess must be positive.\n");
		  printf("Not changing PNSHINE.\n");
		}
	      else
		{
		  new_PN_shininess[0] = newvalue;
		}
	    }


	  // Colour settings for the secondary nodes.
	  else if (0 == strcmp(parameter,"SNDRED"))
	    {
	      // We are trying to set the redness of the secondary nodes.
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing SNDRED.\n");
		}
	      else
		{
		  new_SN_diffuse[0] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"SNDGREEN"))
	    {
	      // We are trying to set the greenness of the secondary nodes.
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing SNDGREEN.\n");
		}
	      else
		{
		  new_SN_diffuse[1] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"SNDBLUE"))
	    {
	      // We are trying to set the blueness of the secondary nodes.
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing SNDBLUE.\n");
		}
	      else
		{
		  new_SN_diffuse[2] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"SNDALPHA"))
	    {
	      // We are trying to set the alpha of the secondary nodes.
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing SNDALPHA.\n");
		}
	      else
		{
		  new_SN_diffuse[3] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"SNSRED"))
	    {
	      // We are trying to set the redness of the secondary nodes.
	      // (specular)
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing SNSRED.\n");
		}
	      else
		{
		  new_SN_specular[0] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"SNSGREEN"))
	    {
	      // We are trying to set the greenness of the secondary nodes.
	      // (specular)
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing SNSGREEN.\n");
		}
	      else
		{
		  new_SN_specular[1] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"SNSBLUE"))
	    {
	      // We are trying to set the blueness of the secondary nodes.
	      // (specular)
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing SNSBLUE.\n");
		}
	      else
		{
		  new_SN_specular[2] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"SNSALPHA"))
	    {
	      // We are trying to set the alpha of the secondary nodes.
	      // (specular)
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing SNSALPHA.\n");
		}
	      else
		{
		  new_SN_specular[3] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"SNSHINE"))
	    {
	      // We are trying to set the shininess of the secondary nodes.
	      // (specular)
	      // Check that it is positive.
	      if (0 >= newvalue)
		{
		  printf("Shininess must be positive.\n");
		  printf("Not changing SNSHINE.\n");
		}
	      else
		{
		  new_SN_shininess[0] = newvalue;
		}
	    }


	  // Colour settings for the primary bonds.
	  else if (0 == strcmp(parameter,"PBDRED"))
	    {
	      // We are trying to set the redness of the primary bonds.
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing PBDRED.\n");
		}
	      else
		{
		  new_PB_diffuse[0] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"PBDGREEN"))
	    {
	      // We are trying to set the greenness of the primary bonds.
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing PBDGREEN.\n");
		}
	      else
		{
		  new_PB_diffuse[1] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"PBDBLUE"))
	    {
	      // We are trying to set the blueness of the primary bonds.
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing PBDBLUE.\n");
		}
	      else
		{
		  new_PB_diffuse[2] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"PBDALPHA"))
	    {
	      // We are trying to set the alpha of the primary bonds.
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing PBDALPHA.\n");
		}
	      else
		{
		  new_PB_diffuse[3] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"PBSRED"))
	    {
	      // We are trying to set the redness of the primary bonds.
	      // (specular)
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing PBSRED.\n");
		}
	      else
		{
		  new_PB_specular[0] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"PBSGREEN"))
	    {
	      // We are trying to set the greenness of the primary bonds.
	      // (specular)
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing PBSGREEN.\n");
		}
	      else
		{
		  new_PB_specular[1] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"PBSBLUE"))
	    {
	      // We are trying to set the blueness of the primary bonds.
	      // (specular)
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing PBSBLUE.\n");
		}
	      else
		{
		  new_PB_specular[2] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"PBSALPHA"))
	    {
	      // We are trying to set the alpha of the primary bonds.
	      // (specular)
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing PBSALPHA.\n");
		}
	      else
		{
		  new_PB_specular[3] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"PBSHINE"))
	    {
	      // We are trying to set the shininess of the primary bonds.
	      // (specular)
	      // Check that it is positive.
	      if (0 >= newvalue)
		{
		  printf("Shininess must be positive.\n");
		  printf("Not changing PBSHINE.\n");
		}
	      else
		{
		  new_PB_shininess[0] = newvalue;
		}
	    }


	  // Colour settings for the secondary bonds.
	  else if (0 == strcmp(parameter,"SBDRED"))
	    {
	      // We are trying to set the redness of the secondary bonds.
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing SBDRED.\n");
		}
	      else
		{
		  new_SB_diffuse[0] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"SBDGREEN"))
	    {
	      // We are trying to set the greenness of the secondary bonds.
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing SBDGREEN.\n");
		}
	      else
		{
		  new_SB_diffuse[1] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"SBDBLUE"))
	    {
	      // We are trying to set the blueness of the secondary bonds.
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing SBDBLUE.\n");
		}
	      else
		{
		  new_SB_diffuse[2] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"SBDALPHA"))
	    {
	      // We are trying to set the alpha of the secondary bonds.
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing SBDALPHA.\n");
		}
	      else
		{
		  new_SB_diffuse[3] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"SBSRED"))
	    {
	      // We are trying to set the redness of the secondary bonds.
	      // (specular)
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing SBSRED.\n");
		}
	      else
		{
		  new_SB_specular[0] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"SBSGREEN"))
	    {
	      // We are trying to set the greenness of the secondary bonds.
	      // (specular)
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing SBSGREEN.\n");
		}
	      else
		{
		  new_SB_specular[1] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"SBSBLUE"))
	    {
	      // We are trying to set the blueness of the secondary bonds.
	      // (specular)
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing SBSBLUE.\n");
		}
	      else
		{
		  new_SB_specular[2] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"SBSALPHA"))
	    {
	      // We are trying to set the alpha of the secondary bonds.
	      // (specular)
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing SBSALPHA.\n");
		}
	      else
		{
		  new_SB_specular[3] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"SBSHINE"))
	    {
	      // We are trying to set the shininess of the secondary bonds.
	      // (specular)
	      // Check that it is positive.
	      if (0 >= newvalue)
		{
		  printf("Shininess must be positive.\n");
		  printf("Not changing SBSHINE.\n");
		}
	      else
		{
		  new_SB_shininess[0] = newvalue;
		}
	    }

	}
      fclose(settings);
    }
  
  // Now set our variables.
  bondThickness = newBondThickness;
  nodeSize = newNodeSize;
  spinAmount = newSpinAmount;
  moveAmount = newMoveAmount;
  scaleAmount = newScaleAmount;

  lightStyle = newLightStyle;
  memcpy(light_position, new_light_position, 4*sizeof(GLfloat));
  myPositionX = newPositionX;
  myPositionY = newPositionY;
  myPositionZ = newPositionZ;

  twistThreshold = newTwistThreshold;
  twistAmount = newTwistAmount;

  axisMarkerSize = newAxisMarkerSize;
  axisLineSize = newAxisLineSize;
  memcpy(XAxis_diffuse, new_XAxis_diffuse, 4*sizeof(GLfloat));
  memcpy(YAxis_diffuse, new_YAxis_diffuse, 4*sizeof(GLfloat));
  memcpy(ZAxis_diffuse, new_ZAxis_diffuse, 4*sizeof(GLfloat));

  memcpy(prim_node_diffuse, new_PN_diffuse, 4*sizeof(GLfloat));
  memcpy(prim_node_specular, new_PN_specular, 4*sizeof(GLfloat));
  memcpy(prim_node_shininess, new_PN_shininess, 1*sizeof(GLfloat));

  memcpy(sec_node_diffuse, new_SN_diffuse, 4*sizeof(GLfloat));
  memcpy(sec_node_specular, new_SN_specular, 4*sizeof(GLfloat));
  memcpy(sec_node_shininess, new_SN_shininess, 1*sizeof(GLfloat));

  memcpy(prim_bond_diffuse, new_PB_diffuse, 4*sizeof(GLfloat));
  memcpy(prim_bond_specular, new_PB_specular, 4*sizeof(GLfloat));
  memcpy(prim_bond_shininess, new_PB_shininess, 1*sizeof(GLfloat));

  memcpy(sec_bond_diffuse, new_SB_diffuse, 4*sizeof(GLfloat));
  memcpy(sec_bond_specular, new_SB_specular, 4*sizeof(GLfloat));
  memcpy(sec_bond_shininess, new_SB_shininess, 1*sizeof(GLfloat));  
}


// The function that does the displaying.
void display()
{
  // Counters, etc.
  int i,j,k;

  int size = cluster->size;


  // Some coordinates to help us keep track of where we are.
  // We will put the minimum x, y, and z here, and then 
  // increment by ones.
  double minCoord = -(double)(size-1)/(double)2;

  // Clear the screen to black (or colour specified in init).
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  // We set up our perspective: 
  // looking towards the origin, and (0,1,0) is up.
  //gluLookAt(0.0, 0.0, myPosition,
  //    0.0, 0.0, 0.0, 0.0, 1.0, 0.0);
  gluLookAt(myPositionX, myPositionY, myPositionZ,
	    0.0, 0.0, 0.0, 0.0, 1.0, 0.0);

  // Set up the light right where we are.
  glLightfv(GL_LIGHT0, GL_POSITION, light_position);

  // Now we go about the actual drawing.
  // We go through the array cluster->adj, and if it's > 0
  // we draw a node.

  if (showAbsAxes == 1)
    {
      // Mark the absolute positive X, Y, and Z axes.
      glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, XAxis_diffuse);
      glPushMatrix();
      glTranslated(1.0*(double)size, 0, 0);
      glRotated(270.0, 0.0, 0.0, 1.0);
      glScaled(axisMarkerSize, axisMarkerSize, axisMarkerSize);
      glutSolidIcosahedron();
      glPopMatrix();
      
      glPushMatrix();
      glTranslated(0.5*(double)size, 0, 0);
      glScaled(1.0*(double)size*(1.0/axisLineSize), 1, 1);
      glutSolidCube(axisLineSize);
      glPopMatrix();
      
      glPushMatrix();
      glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, YAxis_diffuse);
      glTranslated(0, 1.0*(double)size, 0);
      glScaled(axisMarkerSize, axisMarkerSize, axisMarkerSize);
      glutSolidIcosahedron();
      glPopMatrix();
      
      glPushMatrix();
      glTranslated(0, 0.5*(double)size, 0);
      glScaled(1, 1.0*(double)size*(1.0/axisLineSize), 1);
      glutSolidCube(axisLineSize);
      glPopMatrix();
      
      glPushMatrix();
      glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, ZAxis_diffuse);
      glTranslated(0, 0, 1.0*(double)size);
      glRotated(90.0, 1.0, 0.0, 0.0);
      glScaled(axisMarkerSize, axisMarkerSize, axisMarkerSize);
      glutSolidIcosahedron();
      glPopMatrix();

      glPushMatrix();
      glTranslated(0, 0, 0.5*(double)size);
      glScaled(1, 1, 1.0*(double)size*(1.0/axisLineSize));
      glutSolidCube(axisLineSize);
      glPopMatrix();
      
    }

  // Set up the translations first.
  glTranslated(moveX,moveY,moveZ);

  // Set up the rotations second.
  glRotated(theta, 1.0, 0.0, 0.0);
  glRotated(phi, 0.0, 1.0, 0.0);
  glRotated(psi, 0.0, 0.0, 1.0);
  
  //spinMe();

  // Now the scaling.
  glScaled(scale,scale,scale);


  if (showLocalAxes == 1)
    {
      // Mark the axes of the coordinate system local to the
      // cluster (useful for rotations when we get them to work.)
      glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, XAxis_diffuse);
      glPushMatrix();
      glTranslated(0.6*(double)size, 0, 0);
      glRotated(270.0, 0.0, 0.0, 1.0);
      glScaled(axisMarkerSize, axisMarkerSize, axisMarkerSize);
      glutSolidOctahedron();
      glPopMatrix();
      
      glPushMatrix();
      glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, YAxis_diffuse);
      glTranslated(0, 0.6*(double)size, 0);
      glScaled(axisMarkerSize, axisMarkerSize, axisMarkerSize);
      glutSolidOctahedron();
      glPopMatrix();
      
      glPushMatrix();
      glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, ZAxis_diffuse);
      glTranslated(0, 0, 0.6*(double)size);
      glRotated(90.0, 1.0, 0.0, 0.0);
      glScaled(axisMarkerSize, axisMarkerSize, axisMarkerSize);
      glutSolidOctahedron();
      glPopMatrix();
    }

  
  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, prim_node_diffuse);
  glMaterialfv(GL_FRONT, GL_SPECULAR, prim_node_specular);
  glMaterialfv(GL_FRONT, GL_SHININESS, prim_node_shininess);

  // This displays the nodes.
  for (i=0; i<size; i++)
    {
      for (j=0; j<size; j++)
	{
	  for (k=0; k<size; k++)
	    {
	      if (0 < (cluster->adj)[i][j][k])
		{
		  // We draw something here.
		  glPushMatrix();
		  glTranslated(minCoord + (double)i,
			       minCoord + (double)j,
			       minCoord + (double)k);
		  //glScalef(nodeSize, nodeSize, nodeSize);
		  //glutSolidTetrahedron();
		  //glutSolidOctahedron();
		  //glutSolidTeapot(nodeSize);
		  glutSolidCube(nodeSize);
		  glPopMatrix();
		}
	    }
	}
    }


  // Now we attempt to show the bonds.
  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, prim_bond_diffuse);
  glMaterialfv(GL_FRONT, GL_SPECULAR, prim_bond_specular);
  glMaterialfv(GL_FRONT, GL_SHININESS, prim_bond_shininess);

  for (i=0; i<size; i++)
    {
      for (j=0; j<size; j++)
	{
	  for (k=0; k<size; k++)
	    {
	      if (0 < (cluster->adj)[i][j][k])
		{
		  // What do we need to draw?
		  if (i<size-1 && 1==(cluster->x)[i][j][k])
		    {	
		      // Draw an elongated cube connecting (i,j,k) to 
		      // (i+1,j,k).
		      glPushMatrix();
		      glTranslatef(minCoord+0.5+(double)i,
				   minCoord+(double)j,
				   minCoord+(double)k);
		      glScalef(1/bondThickness, 1.0, 1.0);
		      glutSolidCube(bondThickness);
		      glPopMatrix();
		    }

		  if (j<size-1 && 1==(cluster->y)[i][j][k])
		    {
		      // Draw an elongated cube connecting (i,j,k) to 
		      // (i,j+1,k).
		      glPushMatrix();
		      glTranslatef(minCoord+(double)i,
				   minCoord+0.5+(double)j,
				   minCoord+(double)k);
		      glScalef(1.0, 1/bondThickness, 1.0);
		      glutSolidCube(bondThickness);
		      glPopMatrix();
		    }

		  if (k<size-1 && 1==(cluster->z)[i][j][k])
		    {
		      // Draw an elongated cube connecting
		      // (i,j,k) to (i,j,k+1).
		      glPushMatrix();
		      glTranslatef(minCoord+(double)i,
				   minCoord+(double)j,
				   minCoord+0.5+(double)k);
		      glScalef(1.0, 1.0, 1/bondThickness);
		      glutSolidCube(bondThickness);
		      glPopMatrix();
		    }
		}
	    }
	}
    }

  
  // If we want all the information.
  if (1 == toShow)
    {
      glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, sec_node_diffuse);
      glMaterialfv(GL_FRONT, GL_SPECULAR, sec_node_specular);
      glMaterialfv(GL_FRONT, GL_SHININESS, sec_node_shininess);
      
      // This displays the nodes.
      for (i=0; i<size; i++)
	{
	  for (j=0; j<size; j++)
	    {
	      for (k=0; k<size; k++)
		{
		  if (0 > (cluster->adj)[i][j][k])
		    {
		      // We draw something here.
		      glPushMatrix();
		      glTranslated(minCoord + (double)i,
				   minCoord + (double)j,
				   minCoord + (double)k);
		      //glScalef(nodeSize, nodeSize, nodeSize);
		      //glutSolidOctahedron();
		      glutSolidCube(nodeSize);
		      glPopMatrix();
		    }
		}
	    }
	}
      
      
      // Now we attempt to show the bonds.
      glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, sec_bond_diffuse);
      glMaterialfv(GL_FRONT, GL_SPECULAR, sec_bond_specular);
      glMaterialfv(GL_FRONT, GL_SHININESS, sec_bond_shininess);
      
      for (i=0; i<size; i++)
	{
	  for (j=0; j<size; j++)
	    {
	      for (k=0; k<size; k++)
		{
		  if (0 > (cluster->adj)[i][j][k])
		    {
		      // What do we need to draw?
		      if (i<size-1 && 1==(cluster->x)[i][j][k])
			{	
			  // Draw an elongated cube connecting (i,j,k) to 
			  // (i+1,j,k).
			  glPushMatrix();
			  glTranslatef(minCoord+0.5+(double)i,
				       minCoord+(double)j,
				       minCoord+(double)k);
			  glScalef(1/bondThickness, 1.0, 1.0);
			  glutSolidCube(bondThickness);
			  glPopMatrix();
			}
		      
		      if (j<size-1 && 1==(cluster->y)[i][j][k])
			{
			  // Draw an elongated cube connecting (i,j,k) to 
			  // (i,j+1,k).
			  glPushMatrix();
			  glTranslatef(minCoord+(double)i,
				       minCoord+0.5+(double)j,
				       minCoord+(double)k);
			  glScalef(1.0, 1/bondThickness, 1.0);
			  glutSolidCube(bondThickness);
			  glPopMatrix();
			}
		      
		      if (k<size-1 && 1==(cluster->z)[i][j][k])
			{
			  // Draw an elongated cube connecting
			  // (i,j,k) to (i,j,k+1).
			  glPushMatrix();
			  glTranslatef(minCoord+(double)i,
				       minCoord+(double)j,
				       minCoord+0.5+(double)k);
			  glScalef(1.0, 1.0, 1/bondThickness);
			  glutSolidCube(bondThickness);
			  glPopMatrix();
			}
		    }
		}
	    }
	}    
    }
  
  glutSwapBuffers();
  //glFlush();
}


// The reshape function.  As per the way the book does it.
void reshape(int w, int h)
{
  // We get the size locally.
  double size = (double)(cluster->size);

  glViewport(0, 0, (GLsizei) w, (GLsizei) h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  if (0 == myProjection)
    gluPerspective(60.0, 1.0, 1.0, 
		   2 * sqrt(myPositionX*myPositionX + myPositionY*myPositionY
			    + myPositionZ*myPositionZ));
  else
    glOrtho(-size, size, -size, size, 0.1, 
	    sqrt(myPositionX*myPositionX + myPositionY*myPositionY +
		 myPositionZ*myPositionZ)*2);
  glMatrixMode(GL_MODELVIEW);
}


// Spins the cluster.
void spinCluster()
{
  if (0 == twistAxis)
    {
      currTwistY = currTwistZ = 0.0;
      twistModeY = twistModeZ = 0;

      if (0 == twistModeX)
	{
	  // This is spinning in the positive direction.
	  theta = theta + twistAmount;
	  currTwistX = currTwistX + twistAmount;
	  if (currTwistX >= twistThreshold)
	    twistModeX = 1;
	}
      else
	{
	  // Spin in the other direction.
	  theta = theta - twistAmount;
	  currTwistX = currTwistX - twistAmount;
	  if (currTwistX <= -twistThreshold)
	    twistModeX = 0;
	}
    }
  else if (1 == twistAxis)
    {
      currTwistX = currTwistZ = 0.0;
      twistModeX = twistModeZ = 0;

      if (0 == twistModeY)
	{
	  // This is spinning in the positive direction.
	  phi = phi + twistAmount;
	  currTwistY = currTwistY + twistAmount;
	  if (currTwistY >= twistThreshold)
	    twistModeY = 1;
	}
      else
	{
	  // Spin in the other direction.
	  phi = phi - twistAmount;
	  currTwistY = currTwistY - twistAmount;
	  if (currTwistY <= -twistThreshold)
	    twistModeY = 0;
	}
    }
  else if (2 == twistAxis)
    {
      currTwistX = currTwistY = 0.0;
      twistModeX = twistModeY = 0;

      if (0 == twistModeZ)
	{
	  // This is spinning in the positive direction.
	  psi = psi + twistAmount;
	  currTwistZ = currTwistZ + twistAmount;
	  if (currTwistZ >= twistThreshold)
	    twistModeZ = 1;
	}
      else
	{
	  // Spin in the other direction.
	  psi = psi - twistAmount;
	  currTwistZ = currTwistZ - twistAmount;
	  if (currTwistZ <= -twistThreshold)
	    twistModeZ = 0;
	}
    }
  
  glutPostRedisplay();
}


// The keyboard function.  Manipulates our cluster through
// theta, phi, and psi.
void keyboard(unsigned char key, int x, int y)
{
  // We get the size of the cluster.
  double size = (double)(cluster->size);
  
  if ('a' == key)
    {
      // Spin it clockwise about the y-axis (looking down).
      phi = phi+spinAmount;
      if (phi > 360.0)
	phi = phi - 360.0;

      currTwistX = currTwistY = currTwistZ = 0.0;
      twistModeX = twistModeY = twistModeZ = 0;

      glutPostRedisplay();
    }
  else if ('d' == key)
    {
      // Spin it counterclockwise about the y-axis (looking down).
      phi = phi-spinAmount;
      if (phi < 0.0)
	phi = phi + 360.0;

      currTwistX = currTwistY = currTwistZ = 0.0;
      twistModeX = twistModeY = twistModeZ = 0;

      glutPostRedisplay();
    }
  else if ('w' == key)
    {
      // Spin it clockwise about the x-axis (looking left).
      theta=theta+spinAmount;
      if (theta > 360.0)
	theta = theta - 360.0;

      currTwistX = currTwistY = currTwistZ = 0.0;
      twistModeX = twistModeY = twistModeZ = 0;

      glutPostRedisplay();
    }
  else if ('s' == key)
    {
      // Spin it counterclockwise about the x-axis
      // (looking down the negative direction).
      theta=theta-spinAmount;
      if (theta < 0.0)
	theta = theta + 360.0;

      currTwistX = currTwistY = currTwistZ = 0.0;
      twistModeX = twistModeY = twistModeZ = 0;

      glutPostRedisplay();
    }
  else if ('q' == key)
    {
      // Spin it counterclockwise about the z-axis
      // (looking in the negative direction).
      psi=psi-spinAmount;
      if (psi < 0.0)
	psi = psi + 360.0;

      currTwistX = currTwistY = currTwistZ = 0.0;
      twistModeX = twistModeY = twistModeZ = 0;

      glutPostRedisplay();
    }
  else if ('e' == key)
    {
      // Spin it clockwise, etc.
      psi=psi+spinAmount;
      if (psi > 360.0)
	psi = psi - 360.0;

      currTwistX = currTwistY = currTwistZ = 0.0;
      twistModeX = twistModeY = twistModeZ = 0;

      glutPostRedisplay();
    }
  else if ('+' == key)
    {
      // Enlarge it by 10 percent.
      scale = scale * scaleAmount;
      glutPostRedisplay();
    }
  else if ('-' == key)
    {
      // Shrink it back down.
      scale = scale * (1.0/scaleAmount);
      glutPostRedisplay();
    }
  else if ('i' == key)
    {
      // Translate it upwards along the y axis
      moveY += moveAmount;
      glutPostRedisplay();
    }
  else if ('k' == key)
    {
      // ... and back down the y axis.
      moveY -= moveAmount;
      glutPostRedisplay();
    }
  else if ('l' == key)
    {
      moveX += moveAmount;
      glutPostRedisplay();
    }
  else if ('j' == key)
    {
      moveX -= moveAmount;
      glutPostRedisplay();
    }
  else if ('u' == key)
    {
      moveZ += moveAmount;
      glutPostRedisplay();
    }
  else if ('o' == key)
    {
      moveZ -= moveAmount;
      glutPostRedisplay();
    }
  else if ('c' == key)
    {
      // Clear all our movements.
      moveX = moveY = moveZ = 0.0;
      theta = phi = psi = 0.0;
      scale = 1.0;
      toShow = 0;

      twisting = 0;
      currTwistX = currTwistY = currTwistZ = 0.0;
      twistModeX = twistModeY = twistModeZ = 0;
      twistAxis = 0;
      glutIdleFunc(NULL);

      showLocalAxes = showAbsAxes = lightStyle = 1;
      light_position[3] = (double)lightStyle;
      glutPostRedisplay();
    }
  else if ('1' == key)
    {
      // Show only the main things.
      toShow++;
      toShow = toShow%2;
      glutPostRedisplay();
    }
  else if ('8' == key)
    {
      // This is twisting in the X-axis.
      twisting++;
      twisting = twisting%2;
      twistAxis = 0;
      
      if (1 == twisting)
	glutIdleFunc(spinCluster);
      else
	glutIdleFunc(NULL);
    }
  else if ('9' == key)
    {
      // This is spinning in the Y-axis.
      twisting++;
      twisting = twisting%2;
      twistAxis = 1;
      
      if (1 == twisting)
	glutIdleFunc(spinCluster);
      else
	glutIdleFunc(NULL);
    }
  else if ('0' == key)
    {
      twisting++;
      twisting = twisting%2;
      twistAxis = 2;
      
      if (1 == twisting)
	glutIdleFunc(spinCluster);
      else
	glutIdleFunc(NULL);
    }
  else if ('2' == key)
    {
      // Toggle displaying the local axes of the cluster.
      showLocalAxes++;
      showLocalAxes = showLocalAxes%2;
      glutPostRedisplay();
    }
  else if ('3' == key)
    {
      // Toggle displaying the absolute axes.
      showAbsAxes++;
      showAbsAxes = showAbsAxes%2;
      glutPostRedisplay();
    }
  else if ('p' == key)
    {
      // Indicate a change in our projection matrix.
      myProjection++;
      myProjection = myProjection%2;

      // Now, we set the projection matrix.
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      if (0 == myProjection)
	gluPerspective(60.0, 1.0, 1.0, 
		       2 * sqrt(myPositionX*myPositionX 
				+ myPositionY*myPositionY
				+ myPositionZ*myPositionZ));
      else
	glOrtho(-size, size, -size, size, 0.1, 
		sqrt(myPositionX*myPositionX + myPositionY*myPositionY +
		     myPositionZ*myPositionZ)*2);
      glMatrixMode(GL_MODELVIEW);
      
      glutPostRedisplay();
    }
  else if ('m' == key)
    {
      // This is a change in our lighting.
      lightStyle++;
      lightStyle = lightStyle%2;
      light_position[3] = (double)lightStyle;
      glutPostRedisplay();
    }
  else if ('~' == key)
    {
      // First free the cluster.
      dispose_percCluster3D_t(cluster);
      // Exit.
      exit(0);
    }
}
