/* Richard Liang
   Routines that take a percolation cluster and draws it.

   June 10, 2003
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <GL/glut.h>
#include "walk_functions.h"
#include "show_walk.h"


// Global variables (ick) so that we can enable the display
// function to see the percolation cluster.
walk_t *walk;

// And more shared variables to do with manipulating the
// walk so we can play with it.
// these variables rotate in the x, y, and z axes respectively.
double theta;
double phi;
double psi;
double spinAmount;


// Indicates if we set the thing on autopilot twisting.
int twisting;
// The threshold for the twisting: how many degrees we go either way.
int twistThreshold;
// The current twist positions for all three axes.
double currTwistX, currTwistY, currTwistZ;
// How fast we twist.
double twistAmount;
// Which directions we are currently twisting (clockwise or counterclockwise)
int twistModeX, twistModeY, twistModeZ;
// Which axis we are currently twisting in (0~x, 1~y, 2~z)
int twistAxis;


// These variables translate the walk.
double moveX;
double moveY;
double moveZ;
double moveAmount;

// This variable scales the walk.
double scale;
double scaleAmount;


// This variable sets the viewer's position.
double myPositionX;
double myPositionY;
double myPositionZ;

// This variable determines whether we want to use
// an orthogonal or perspective projection matrix.
// 0 for perspective, 1 for orthogonal.
int myProjection;

// Particulars for visual tuning.
float pathThickness;
float nodeSize;
float originSize;
int lightStyle;
GLfloat light_position[4];

// Axis markers.
float axisMarkerSize;
float axisLineSize;
GLfloat XAxis_diffuse[4];
GLfloat YAxis_diffuse[4];
GLfloat ZAxis_diffuse[4];

// Indicates whether or not we are going to show the 
// axes of the absolute coordinate system.
int showAbsAxes;

// Indicates whether or not we are going to show the
// axes of the coordinate system local to the structure.
int showLocalAxes;


// The colours we use.
GLfloat node_diffuse[4];
GLfloat node_specular[4];
GLfloat node_shininess[1];

GLfloat path_diffuse[4];
GLfloat path_specular[4];
GLfloat path_shininess[1];

GLfloat origin_diffuse[4];
GLfloat origin_specular[4];
GLfloat origin_shininess[1];


// The name of our config file.
char configfile[50] = "randWalk.conf";


// This function does the drawing... It is like the "main"
// function specified in the book.
void drawWalk(walk_t *myWalk,
		 int *argc, char **argv)
{
  // Mostly this is lifted from examples in the book.
  glutInit(argc, argv);
  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
  glutInitWindowSize(500,500);
  glutInitWindowPosition(50,50);
  glutCreateWindow("My Random Walk");

  // Here we set up the clear colour, etc.
  walk = myWalk;
  init();

  glutDisplayFunc(display);
  glutKeyboardFunc(keyboard);
  glutReshapeFunc(reshape);
  glutMainLoop();
}

// The init function.  This sets all the things we only set
// once during the lifetime of the display.
void init()
{
  // Information regarding the light: we set one light, 
  // right where we are, as if we are using a flashlight.
  // This is mostly copied from p.180 of the book.
  //GLfloat light_position[] = {0.0, 0.0, myPosition, 1.0};
  GLfloat white_light[] = {1.0, 1.0, 1.0, 1.0};
  GLfloat lmodel_ambient[] = {0.1, 0.1, 0.1, 1.0};

  // Set the clear color.
  glClearColor(0.0, 0.0, 0.0, 0.0);
  // We don't want any fancy-pants colouring: all polygons
  // will be uniformly coloured.
  glShadeModel(GL_FLAT);

  theta = 0.0;
  phi = 0.0;
  psi = 0.0;
  
  moveX = 0.0;
  moveY = 0.0;
  moveZ = 0.0;

  scale = 1.0;

  // Twisting stuff.
  twisting = 0;
  currTwistX = currTwistY = currTwistZ = 0.0;
  twistModeX = twistModeY = twistModeZ = 0;
  twistAxis = 0;

  showAbsAxes = showLocalAxes = 1;

  // Read some preferences from our .conf file.
  config();

  // Indicate that initially we are to use a perspective 
  // projection matrix.
  myProjection = 0;

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  // Set lighting up.  We set all material things to have
  // the same properties.
  //glLightfv(GL_LIGHT0, GL_POSITION, light_position);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, white_light);
  glLightfv(GL_LIGHT0, GL_SPECULAR, white_light);
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);

  // We want to attenuate the light at a distance.
  //glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION, 1.5);
  //glLightf(GL_LIGHT0, GL_LINEAR_ATTENUATION, 0.5);
  //glLightf(GL_LIGHT0, GL_QUADRATIC_ATTENUATION, 0.2);

  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glEnable(GL_DEPTH_TEST);


  // Things seem to get darker as we scale.  This should fix that.
  //glEnable(GL_RESCALE_NORMAL);
  // That worked, but somehow the z-connectors weren't the right
  // colour.  This fixes it (why?).
  glEnable(GL_NORMALIZE);


  // Prettying things up a bit more with antialiasing.
  // Most of this is taken from the book.
  //glEnable(GL_POLYGON_SMOOTH);
  //glEnable(GL_BLEND);
  //glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  //glHint(GL_POLYGON_SMOOTH_HINT, GL_DONT_CARE);

}


// A function to read configuration settings from a configuration file.
// Note: don't be fooled, this is a very simple function and all it
// does is read in settings from a file.  Unfortunately it's also
// quite a long function.
void config()
{
  // The size of the walk.
  double size = 2*(double)(walk->maxSize);
  // A filehandle.
  FILE *settings;

  // Things we will set.  These are the default values.
  double newPathThickness = 0.10;
  double newNodeSize = 0.10;
  double newOriginSize = 0.20;
  double newSpinAmount = 1.0;
  double newMoveAmount = 0.01;
  double newScaleAmount = 1.1;


  double newAxisMarkerSize = 0.2;
  double newAxisLineSize = 0.015;
  GLfloat new_XAxis_diffuse[] = {1.0, 0.0, 0.0, 1.0};
  GLfloat new_YAxis_diffuse[] = {0.0, 1.0, 0.0, 1.0};
  GLfloat new_ZAxis_diffuse[] = {0.0, 0.0, 1.0, 1.0};


  // Camera.
  double newPositionX = 1.2*size;
  double newPositionY = newPositionX;
  double newPositionZ = newPositionX;

  
  // Lights.
  int newLightStyle = 1;
  GLfloat new_light_position[4] = {newPositionX, newPositionY, newPositionZ, 
				   (double)newLightStyle};

  
  // Twisting stuff.
  int newTwistThreshold = 40;
  double newTwistAmount = 1.0;

  // Colours.
  GLfloat new_node_diffuse[4] = {1.0, 1.0, 1.0, 1.0};
  GLfloat new_node_specular[4] = {1.0, 1.0, 1.0, 1.0};
  GLfloat new_node_shininess[1] = {120.0};

  GLfloat new_path_diffuse[] = {0.0, 1.0, 0.0, 1.0};
  GLfloat new_path_specular[] = {1.0, 1.0, 1.0, 1.0};
  GLfloat new_path_shininess[] = { 120.0 };

  GLfloat new_origin_diffuse[] = {1.0, 0.0, 0.0, 1.0};
  GLfloat new_origin_specular[] = {1.0, 1.0, 1.0, 1.0};
  GLfloat new_origin_shininess[] = { 120.0 };

  int n;

  char parameter[50] = "";
  double newvalue = -1;
  char del[10] = "";

  // Open the file.
  if (NULL == (settings = fopen(configfile,"r")))
    {
      printf("Could not open the config file.  Using default settings.\n\n");
    }
  else
    {
      // Start reading.
      while ((3 == (n = fscanf(settings,"%s = %lf %1s",parameter,&newvalue,del))) &&
	     (0 == feof(settings)) && (0 == ferror(settings)) && (del[0] == 'd'))
	{
	  if (0 == strcmp(parameter,"PATHTHICKNESS"))
	    {
	      // We are trying to set the path thickness.  Check
	      // that it is positive.
	      if (0 >= newvalue)
		{
		  printf("Path thickness must be positive.\n");
		  printf("Not changing the path thickness.\n");
		}
	      else
		{
		  newPathThickness = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"NODESIZE"))
	    {
	      // We are trying to set node size.  Check that it is positive.
	      if (0 >= newvalue)
		{
		  printf("Node size must be positive.\n");
		  printf("Not changing the node size.\n");
		}
	      else
		{
		  newNodeSize = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"ORIGINSIZE"))
	    {
	      // We are trying to set origin size.  
	      // Check that it is positive.
	      if (0 >= newvalue)
		{
		  printf("Origin size must be positive.\n");
		  printf("Not changing the origin size.\n");
		}
	      else
		{
		  newOriginSize = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"AXISMARKERSIZE"))
	    {
	      // We are trying to set axis marker size.  
	      // Check that it is not negative.
	      if (0 > newvalue)
		{
		  printf("Axis marker size must be non-negative.\n");
		  printf("Not changing the axis marker size.\n");
		}
	      else
		{
		  newAxisMarkerSize = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"AXISLINESIZE"))
	    {
	      // We are trying to set axis line size.  
	      // Check that it is non-negative.
	      if (0 > newvalue)
		{
		  printf("Axis line size must be non-negative.\n");
		  printf("Not changing the axis line size.\n");
		}
	      else
		{
		  newAxisLineSize = newvalue;
		}
	    }	  
	  else if (0 == strcmp(parameter,"MOVEAMOUNT"))
	    {
	      // We are trying to set the move amount.  Check that it is positive.
	      if (0 >= newvalue)
		{
		  printf("Move amount must be positive.\n");
		  printf("Not changing move amount.\n");
		}
	      else
		{
		  newMoveAmount = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"SPINAMOUNT"))
	    {
	      // We are trying to set the spin amount.  Check that it is positive.
	      if (0 >= newvalue)
		{
		  printf("Spin amount must be positive.\n");
		  printf("Not changing spin amount.\n");
		}
	      else
		{
		  newSpinAmount = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"SCALEAMOUNT"))
	    {
	      // We are trying to set the scale amount.  
	      // Check that it is greater than 1.
	      if (1 >= newvalue)
		{
		  printf("Scale amount must be greater than 1.\n");
		  printf("Not changing scale amount.\n");
		}
	      else
		{
		  newScaleAmount = newvalue;
		}
	    }
	  
	  // Twisting stuff.
	  else if (0 == strcmp(parameter,"TWISTAMOUNT"))
	    {
	      // We are trying to set the twist amount.  
	      // Check that it is positive.
	      if (0 >= newvalue)
		{
		  printf("Twist amount must be positive.\n");
		  printf("Not changing twist amount.\n");
		}
	      else
		{
		  newTwistAmount = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"THRESHOLD"))
	    {
	      // We are trying to set the threshold.  
	      // Check that it is greater than 0.
	      if (0 >= newvalue)
		{
		  printf("Twist threshold must be greater than 1.\n");
		  printf("Not changing twist threshold.\n");
		}
	      else
		{
		  newTwistThreshold = (int)newvalue;
		}
	    }


	  // Camera settings.
	  else if (0 == strcmp(parameter,"CAMERAX"))
	    {
	      newPositionX = newvalue * size;
	    }
	  else if (0 == strcmp(parameter,"CAMERAY"))
	    {
	      newPositionY = newvalue * size;
	    }
	  else if (0 == strcmp(parameter,"CAMERAZ"))
	    {
	      newPositionZ = newvalue * size;
	    }

	  // Light settings.
	  else if (0 == strcmp(parameter,"LIGHTX"))
	    {
	      new_light_position[0] = newvalue * size;
	    }
	  else if (0 == strcmp(parameter,"LIGHTY"))
	    {
	      new_light_position[1] = newvalue * size;
	    }
	  else if (0 == strcmp(parameter,"LIGHTZ"))
	    {
	      new_light_position[2] = newvalue * size;
	    }
	    



	  // Colour settings for the nodes.
	  else if (0 == strcmp(parameter,"NDRED"))
	    {
	      // We are trying to set the redness of the nodes.
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing NDRED.\n");
		}
	      else
		{
		  new_node_diffuse[0] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"NDGREEN"))
	    {
	      // We are trying to set the greenness of the nodes.
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing NDGREEN.\n");
		}
	      else
		{
		  new_node_diffuse[1] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"NDBLUE"))
	    {
	      // We are trying to set the blueness of the nodes.
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing NDBLUE.\n");
		}
	      else
		{
		  new_node_diffuse[2] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"NDALPHA"))
	    {
	      // We are trying to set the alpha of the nodes.
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing NDALPHA.\n");
		}
	      else
		{
		  new_node_diffuse[3] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"NSRED"))
	    {
	      // We are trying to set the redness of the nodes.
	      // (specular)
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing NSRED.\n");
		}
	      else
		{
		  new_node_specular[0] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"NSGREEN"))
	    {
	      // We are trying to set the greenness of the nodes.
	      // (specular)
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing NSGREEN.\n");
		}
	      else
		{
		  new_node_specular[1] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"NSBLUE"))
	    {
	      // We are trying to set the blueness of the nodes.
	      // (specular)
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing NSBLUE.\n");
		}
	      else
		{
		  new_node_specular[2] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"NSALPHA"))
	    {
	      // We are trying to set the alpha of the nodes.
	      // (specular)
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing NSALPHA.\n");
		}
	      else
		{
		  new_node_specular[3] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"NSHINE"))
	    {
	      // We are trying to set the shininess of the nodes.
	      // Check that it is positive.
	      if (0 >= newvalue)
		{
		  printf("Shininess must be positive.\n");
		  printf("Not changing NSHINE.\n");
		}
	      else
		{
		  new_node_shininess[0] = newvalue;
		}
	    }


	  // Colour settings for the path.
	  else if (0 == strcmp(parameter,"PDRED"))
	    {
	      // We are trying to set the redness of the path.
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing PDRED.\n");
		}
	      else
		{
		  new_path_diffuse[0] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"PDGREEN"))
	    {
	      // We are trying to set the greenness of the path.
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing PDGREEN.\n");
		}
	      else
		{
		  new_path_diffuse[1] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"PDBLUE"))
	    {
	      // We are trying to set the blueness of the path.
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing PDBLUE.\n");
		}
	      else
		{
		  new_path_diffuse[2] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"PDALPHA"))
	    {
	      // We are trying to set the alpha of the path.
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing PDALPHA.\n");
		}
	      else
		{
		  new_path_diffuse[3] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"PSRED"))
	    {
	      // We are trying to set the redness of the path.
	      // (specular)
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing PSRED.\n");
		}
	      else
		{
		  new_path_specular[0] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"PSGREEN"))
	    {
	      // We are trying to set the greenness of the path.
	      // (specular)
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing PSGREEN.\n");
		}
	      else
		{
		  new_path_specular[1] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"PSBLUE"))
	    {
	      // We are trying to set the blueness of the path.
	      // (specular)
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing PSBLUE.\n");
		}
	      else
		{
		  new_path_specular[2] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"PSALPHA"))
	    {
	      // We are trying to set the alpha of the path.
	      // (specular)
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing PSALPHA.\n");
		}
	      else
		{
		  new_path_specular[3] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"PSHINE"))
	    {
	      // We are trying to set the shininess of the path.
	      // (specular)
	      // Check that it is positive.
	      if (0 >= newvalue)
		{
		  printf("Shininess must be positive.\n");
		  printf("Not changing PSHINE.\n");
		}
	      else
		{
		  new_path_shininess[0] = newvalue;
		}
	    }


	  // Colour settings for the origin.
	  else if (0 == strcmp(parameter,"ODRED"))
	    {
	      // We are trying to set the redness of the origin.
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing ODRED.\n");
		}
	      else
		{
		  new_origin_diffuse[0] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"ODGREEN"))
	    {
	      // We are trying to set the greenness of the origin.
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing ODGREEN.\n");
		}
	      else
		{
		  new_origin_diffuse[1] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"ODBLUE"))
	    {
	      // We are trying to set the blueness of the origin.
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing ODBLUE.\n");
		}
	      else
		{
		  new_origin_diffuse[2] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"ODALPHA"))
	    {
	      // We are trying to set the alpha of the origin.
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing ODALPHA.\n");
		}
	      else
		{
		  new_origin_diffuse[3] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"OSRED"))
	    {
	      // We are trying to set the redness of the origin.
	      // (specular)
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing OSRED.\n");
		}
	      else
		{
		  new_origin_specular[0] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"OSGREEN"))
	    {
	      // We are trying to set the greenness of the origin.
	      // (specular)
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing OSGREEN.\n");
		}
	      else
		{
		  new_origin_specular[1] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"OSBLUE"))
	    {
	      // We are trying to set the blueness of the origin.
	      // (specular)
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing OSBLUE.\n");
		}
	      else
		{
		  new_origin_specular[2] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"OSALPHA"))
	    {
	      // We are trying to set the alpha of the origin.
	      // (specular)
	      // Check that it is between 0 and 1.
	      if (0 > newvalue || 1 < newvalue)
		{
		  printf("Color components must be between 0 and 1.\n");
		  printf("Not changing OSALPHA.\n");
		}
	      else
		{
		  new_origin_specular[3] = newvalue;
		}
	    }
	  else if (0 == strcmp(parameter,"OSHINE"))
	    {
	      // We are trying to set the shininess of the origin.
	      // Check that it is positive.
	      if (0 >= newvalue)
		{
		  printf("Shininess must be positive.\n");
		  printf("Not changing OSHINE.\n");
		}
	      else
		{
		  new_origin_shininess[0] = newvalue;
		}
	    }

	}
      fclose(settings);
    }
  
  // Now set our variables.
  pathThickness = newPathThickness;
  nodeSize = newNodeSize;
  originSize = newOriginSize;
  spinAmount = newSpinAmount;
  moveAmount = newMoveAmount;
  scaleAmount = newScaleAmount;

  lightStyle = newLightStyle;
  memcpy(light_position, new_light_position, 4*sizeof(GLfloat));
  myPositionX = newPositionX;
  myPositionY = newPositionY;
  myPositionZ = newPositionZ;

  twistThreshold = newTwistThreshold;
  twistAmount = newTwistAmount;

  axisMarkerSize = newAxisMarkerSize;
  axisLineSize = newAxisLineSize;
  memcpy(XAxis_diffuse, new_XAxis_diffuse, 4*sizeof(GLfloat));
  memcpy(YAxis_diffuse, new_YAxis_diffuse, 4*sizeof(GLfloat));
  memcpy(ZAxis_diffuse, new_ZAxis_diffuse, 4*sizeof(GLfloat));


  memcpy(node_diffuse, new_node_diffuse, 4*sizeof(GLfloat));
  memcpy(node_specular, new_node_specular, 4*sizeof(GLfloat));
  memcpy(node_shininess, new_node_shininess, 1*sizeof(GLfloat));

  memcpy(path_diffuse, new_path_diffuse, 4*sizeof(GLfloat));
  memcpy(path_specular, new_path_specular, 4*sizeof(GLfloat));
  memcpy(path_shininess, new_path_shininess, 1*sizeof(GLfloat));

  memcpy(origin_diffuse, new_origin_diffuse, 4*sizeof(GLfloat));
  memcpy(origin_specular, new_origin_specular, 4*sizeof(GLfloat));
  memcpy(origin_shininess, new_origin_shininess, 1*sizeof(GLfloat));
}


// The function that does the displaying.
void display()
{
  // Counters, etc.
  int i;

  int size = 2 * walk->maxSize;

  int numSteps = walk->numSteps;

  // Some coordinates to help us keep track of where we are.
  int currX = 0;
  int currY = 0;
  int currZ = 0;
  char currStep = 0;

  // Clear the screen to black (or colour specified in init).
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  // We set up our perspective: 
  // looking towards the origin, and (0,1,0) is up.
  gluLookAt(myPositionX, myPositionY, myPositionZ,
	    0.0, 0.0, 0.0, 0.0, 1.0, 0.0);

  // Set up the light right where we are.
  glLightfv(GL_LIGHT0, GL_POSITION, light_position);

  // Now we go about the actual drawing.
  // We go through the walk step by step, drawing cubes
  // at every stop, and using rectangular prisms to show
  // our path.

  if (showAbsAxes == 1)
    {
      // Mark the absolute positive X, Y, and Z axes.
      glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, XAxis_diffuse);
      glPushMatrix();
      glTranslated(1.0*(double)size, 0, 0);
      glRotated(270.0, 0.0, 0.0, 1.0);
      glScaled(axisMarkerSize, axisMarkerSize, axisMarkerSize);
      glutSolidIcosahedron();
      glPopMatrix();
      
      glPushMatrix();
      glTranslated(0.5*(double)size, 0, 0);
      glScaled(1.0*(double)size*(1.0/axisLineSize), 1, 1);
      glutSolidCube(axisLineSize);
      glPopMatrix();
      
      glPushMatrix();
      glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, YAxis_diffuse);
      glTranslated(0, 1.0*(double)size, 0);
      glScaled(axisMarkerSize, axisMarkerSize, axisMarkerSize);
      glutSolidIcosahedron();
      glPopMatrix();
      
      glPushMatrix();
      glTranslated(0, 0.5*(double)size, 0);
      glScaled(1, 1.0*(double)size*(1.0/axisLineSize), 1);
      glutSolidCube(axisLineSize);
      glPopMatrix();
      
      glPushMatrix();
      glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, ZAxis_diffuse);
      glTranslated(0, 0, 1.0*(double)size);
      glRotated(90.0, 1.0, 0.0, 0.0);
      glScaled(axisMarkerSize, axisMarkerSize, axisMarkerSize);
      glutSolidIcosahedron();
      glPopMatrix();

      glPushMatrix();
      glTranslated(0, 0, 0.5*(double)size);
      glScaled(1, 1, 1.0*(double)size*(1.0/axisLineSize));
      glutSolidCube(axisLineSize);
      glPopMatrix();
      
    }


  // Set up the translations first.
  glTranslated(moveX,moveY,moveZ);

  // Set up the rotations second.
  glRotated(theta, 1.0, 0.0, 0.0);
  glRotated(phi, 0.0, 1.0, 0.0);
  glRotated(psi, 0.0, 0.0, 1.0);

  // Now the scaling.
  glScaled(scale,scale,scale);

  
  if (showLocalAxes == 1)
    {
      // Mark the axes of the coordinate system local to the
      // cluster (useful for rotations when we get them to work.)
      glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, XAxis_diffuse);
      glPushMatrix();
      glTranslated(0.6*(double)size, 0, 0);
      glRotated(270.0, 0.0, 0.0, 1.0);
      glScaled(axisMarkerSize, axisMarkerSize, axisMarkerSize);
      glutSolidOctahedron();
      glPopMatrix();
      
      glPushMatrix();
      glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, YAxis_diffuse);
      glTranslated(0, 0.6*(double)size, 0);
      glScaled(axisMarkerSize, axisMarkerSize, axisMarkerSize);
      glutSolidOctahedron();
      glPopMatrix();
      
      glPushMatrix();
      glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, ZAxis_diffuse);
      glTranslated(0, 0, 0.6*(double)size);
      glRotated(90.0, 1.0, 0.0, 0.0);
      glScaled(axisMarkerSize, axisMarkerSize, axisMarkerSize);
      glutSolidOctahedron();
      glPopMatrix();
    }

      
  // Draw the walk.  The first stop we make is at the origin.
  // For orientation purposes, this one can be set different.
  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, origin_diffuse);
  glMaterialfv(GL_FRONT, GL_SPECULAR, origin_specular);
  glMaterialfv(GL_FRONT, GL_SHININESS, origin_shininess);

  glutSolidCube(originSize);
  
  // From now on we draw a path to connect each stop, and then
  // draw the stop.
  for (i=0; i<numSteps; i++)
    {
      currStep = (walk->steps)[i];

      // Display the next step of the path: use the correct colour.
      glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, path_diffuse);
      glMaterialfv(GL_FRONT, GL_SPECULAR, path_specular);
      glMaterialfv(GL_FRONT, GL_SHININESS, path_shininess);
      
      glPushMatrix();
      if (1 == currStep)
	{
	  // We connect (currX,currY,currZ) to
	  // (currX+1,currY,currZ).
	  glTranslated(currX+0.5, currY, currZ);
	  glScaled((double)1/pathThickness, 1, 1);
	  glutSolidCube(pathThickness);
	  currX++;
	}
      else if (-1 == currStep)
	{
	  // We connect (currX,currY,currZ) to
	  // (currX-1,currY,currZ).
	  glTranslated(currX-0.5, currY, currZ);
	  glScaled((double)1/pathThickness, 1.0, 1.0);
	  glutSolidCube(pathThickness);
	  currX--;
	}
      else if (2 == currStep)
	{
	  // We connect (currX,currY,currZ) to
	  // (currX,currY+1,currZ).
	  glTranslated(currX, currY+0.5, currZ);
	  glScaled(1.0, (double)1/pathThickness, 1.0);
	  glutSolidCube(pathThickness);
	  currY++;
	}
      else if (-2 == currStep)
	{
	  // We connect (currX,currY,currZ) to
	  // (currX,currY-1,currZ).
	  glTranslated(currX, currY-0.5, currZ);
	  glScaled(1.0, (double)1/pathThickness, 1.0);
	  glutSolidCube(pathThickness);
	  currY--;
	}
      else if (3 == currStep)
	{
	  // We connect (currX,currY,currZ) to
	  // (currX,currY,currZ+1).
	  glTranslated(currX, currY, currZ+0.5);
	  glScaled(1.0, 1.0, (double)1/pathThickness);
	  glutSolidCube(pathThickness);
	  currZ++;
	}
      else if (-3 == currStep)
	{
	  // We connect (currX,currY,currZ) to
	  // (currX,currY,currZ-1).
	  glTranslated(currX, currY, currZ-0.5);
	  glScaled(1.0, 1.0, (double)1/pathThickness);
	  glutSolidCube(pathThickness);
	  currZ--;
	}
      glPopMatrix();

      glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, node_diffuse);
      glMaterialfv(GL_FRONT, GL_SPECULAR, node_specular);
      glMaterialfv(GL_FRONT, GL_SHININESS, node_shininess);

      // Draw the next stop.
      glPushMatrix();
      glTranslated(currX, currY, currZ);
      glutSolidCube(nodeSize);
      glPopMatrix();
    }
  
  glutSwapBuffers();
}


// The reshape function.  As per the way the book does it.
void reshape(int w, int h)
{
  // We get the size locally.
  double maxSize = 2 * (double)(walk->maxSize);

  glViewport(0, 0, (GLsizei) w, (GLsizei) h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  if (0 == myProjection)
    gluPerspective(60.0, 1.0, 1.0, 
		   2 * sqrt(myPositionX*myPositionX + myPositionY*myPositionY
			    + myPositionZ*myPositionZ));
  else
    glOrtho(-maxSize, maxSize, -maxSize, maxSize, 0.1, 
	    2 * sqrt(myPositionX*myPositionX + myPositionY*myPositionY
		     + myPositionZ*myPositionZ));
  glMatrixMode(GL_MODELVIEW);
}


// Spins the walk.
void spinWalk()
{
  if (0 == twistAxis)
    {
      currTwistY = currTwistZ = 0.0;
      twistModeY = twistModeZ = 0;

      if (0 == twistModeX)
	{
	  // This is spinning in the positive direction.
	  theta = theta + twistAmount;
	  currTwistX = currTwistX + twistAmount;
	  if (currTwistX >= twistThreshold)
	    twistModeX = 1;
	}
      else
	{
	  // Spin in the other direction.
	  theta = theta - twistAmount;
	  currTwistX = currTwistX - twistAmount;
	  if (currTwistX <= -twistThreshold)
	    twistModeX = 0;
	}
    }
  else if (1 == twistAxis)
    {
      currTwistX = currTwistZ = 0.0;
      twistModeX = twistModeZ = 0;

      if (0 == twistModeY)
	{
	  // This is spinning in the positive direction.
	  phi = phi + twistAmount;
	  currTwistY = currTwistY + twistAmount;
	  if (currTwistY >= twistThreshold)
	    twistModeY = 1;
	}
      else
	{
	  // Spin in the other direction.
	  phi = phi - twistAmount;
	  currTwistY = currTwistY - twistAmount;
	  if (currTwistY <= -twistThreshold)
	    twistModeY = 0;
	}
    }
  else if (2 == twistAxis)
    {
      currTwistX = currTwistY = 0.0;
      twistModeX = twistModeY = 0;

      if (0 == twistModeZ)
	{
	  // This is spinning in the positive direction.
	  psi = psi + twistAmount;
	  currTwistZ = currTwistZ + twistAmount;
	  if (currTwistZ >= twistThreshold)
	    twistModeZ = 1;
	}
      else
	{
	  // Spin in the other direction.
	  psi = psi - twistAmount;
	  currTwistZ = currTwistZ - twistAmount;
	  if (currTwistZ <= -twistThreshold)
	    twistModeZ = 0;
	}
    }
  
  glutPostRedisplay();
}



// The keyboard function.  Manipulates our walk in different ways.
void keyboard(unsigned char key, int x, int y)
{
  // We get the size of the walk.
  double size = 2 * (double)(walk->maxSize);
  
  if ('a' == key)
    {
      // Spin it clockwise about the y-axis (looking down).
      phi = phi+spinAmount;
      if (phi > 360.0)
	phi = phi - 360.0;

      currTwistX = currTwistY = currTwistZ = 0.0;
      twistModeX = twistModeY = twistModeZ = 0;

      glutPostRedisplay();
    }
  else if ('d' == key)
    {
      // Spin it counterclockwise about the y-axis (looking down).
      phi = phi-spinAmount;
      if (phi < 0.0)
	phi = phi + 360.0;

      currTwistX = currTwistY = currTwistZ = 0.0;
      twistModeX = twistModeY = twistModeZ = 0;

      glutPostRedisplay();
    }
  else if ('w' == key)
    {
      // Spin it clockwise about the x-axis (looking left).
      theta=theta+spinAmount;
      if (theta > 360.0)
	theta = theta - 360.0;

      currTwistX = currTwistY = currTwistZ = 0.0;
      twistModeX = twistModeY = twistModeZ = 0;

      glutPostRedisplay();
    }
  else if ('s' == key)
    {
      // Spin it counterclockwise about the x-axis
      // (looking down the negative direction).
      theta=theta-spinAmount;
      if (theta < 0.0)
	theta = theta + 360.0;

      currTwistX = currTwistY = currTwistZ = 0.0;
      twistModeX = twistModeY = twistModeZ = 0;

      glutPostRedisplay();
    }
  else if ('q' == key)
    {
      // Spin it counterclockwise about the z-axis
      // (looking in the negative direction).
      psi=psi-spinAmount;
      if (psi < 0.0)
	psi = psi + 360.0;

      currTwistX = currTwistY = currTwistZ = 0.0;
      twistModeX = twistModeY = twistModeZ = 0;

      glutPostRedisplay();
    }
  else if ('e' == key)
    {
      // Spin it clockwise, etc.
      psi=psi+spinAmount;
      if (psi > 360.0)
	psi = psi - 360.0;

      currTwistX = currTwistY = currTwistZ = 0.0;
      twistModeX = twistModeY = twistModeZ = 0;

      glutPostRedisplay();
    }
  else if ('+' == key)
    {
      // Enlarge it by 10 percent.
      scale = scale * scaleAmount;
      glutPostRedisplay();
    }
  else if ('-' == key)
    {
      // Shrink it back down.
      scale = scale * (1.0/scaleAmount);
      glutPostRedisplay();
    }
  else if ('i' == key)
    {
      // Translate it upwards along the y axis
      moveY += moveAmount;
      glutPostRedisplay();
    }
  else if ('k' == key)
    {
      // ... and back down the y axis.
      moveY -= moveAmount;
      glutPostRedisplay();
    }
  else if ('l' == key)
    {
      moveX += moveAmount;
      glutPostRedisplay();
    }
  else if ('j' == key)
    {
      moveX -= moveAmount;
      glutPostRedisplay();
    }
  else if ('u' == key)
    {
      moveZ += moveAmount;
      glutPostRedisplay();
    }
  else if ('o' == key)
    {
      moveZ -= moveAmount;
      glutPostRedisplay();
    }
  else if ('c' == key)
    {
      // Clear all our movements.
      moveX = moveY = moveZ = 0.0;
      theta = phi = psi = 0.0;
      scale = 1.0;

      twisting = 0;
      currTwistX = currTwistY = currTwistZ = 0.0;
      twistModeX = twistModeY = twistModeZ = 0;
      twistAxis = 0;
      glutIdleFunc(NULL);

      showLocalAxes = showAbsAxes = lightStyle = 1;
      light_position[3] = (double)lightStyle;
      glutPostRedisplay();
    }
  else if ('8' == key)
    {
      // This is twisting in the X-axis.
      twisting++;
      twisting = twisting%2;
      twistAxis = 0;
      
      if (1 == twisting)
	glutIdleFunc(spinWalk);
      else
	glutIdleFunc(NULL);
    }
  else if ('9' == key)
    {
      // This is spinning in the Y-axis.
      twisting++;
      twisting = twisting%2;
      twistAxis = 1;
      
      if (1 == twisting)
	glutIdleFunc(spinWalk);
      else
	glutIdleFunc(NULL);
    }
  else if ('0' == key)
    {
      twisting++;
      twisting = twisting%2;
      twistAxis = 2;
      
      if (1 == twisting)
	glutIdleFunc(spinWalk);
      else
	glutIdleFunc(NULL);
    }
  else if ('2' == key)
    {
      // Toggle displaying the local axes of the cluster.
      showLocalAxes++;
      showLocalAxes = showLocalAxes%2;
      glutPostRedisplay();
    }
  else if ('3' == key)
    {
      // Toggle displaying the absolute axes.
      showAbsAxes++;
      showAbsAxes = showAbsAxes%2;
      glutPostRedisplay();
    }
  else if ('p' == key)
    {
      // Indicate a change in our projection matrix.
      myProjection++;
      myProjection = myProjection%2;

      // Now, we set the projection matrix.
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      if (0 == myProjection)
	gluPerspective(60.0, 1.0, 1.0, 
		       2 * sqrt(myPositionX*myPositionX 
				+ myPositionY*myPositionY
				+ myPositionZ*myPositionZ));
      else
	glOrtho(-size, size, -size, size, 1.0, 
		sqrt(myPositionX*myPositionX + myPositionY*myPositionY +
		     myPositionZ*myPositionZ)*2);
      glMatrixMode(GL_MODELVIEW);

      glutPostRedisplay();
    }
  else if ('m' == key)
    {
      // This is a change in our lighting.
      lightStyle++;
      lightStyle = lightStyle%2;
      light_position[3] = (double)lightStyle;
      glutPostRedisplay();
    }
  else if ('~' == key)
    {
      // Exit.
      exit(0);
    }
}
